@@ -215,23 +215,16 @@ void QQmlPrivate::qdeclarativeelement_destructor(QObject *o)
 {
     QObjectPrivate *p = QObjectPrivate::get(o);
     if (QQmlData *d = QQmlData::get(p)) {
+        const auto invalidate = [](QQmlContextData *c) {c->invalidate();};
         if (d->ownContext) {
-            for (QQmlRefPointer<QQmlContextData> lc = d->ownContext->linkedContext(); lc;
-                 lc = lc->linkedContext()) {
-                lc->invalidate();
-                if (lc->contextObject() == o)
-                    lc->setContextObject(nullptr);
-            }
-            d->ownContext->invalidate();
-            if (d->ownContext->contextObject() == o)
-                d->ownContext->setContextObject(nullptr);
+            d->ownContext->deepClearContextObject(o, invalidate, invalidate);
             d->ownContext.reset();
             d->context = nullptr;
+            Q_ASSERT(!d->outerContext || d->outerContext->contextObject() != o);
+        } else if (d->outerContext && d->outerContext->contextObject() == o) {
+            d->outerContext->deepClearContextObject(o, invalidate, invalidate);
         }
 
-        if (d->outerContext && d->outerContext->contextObject() == o)
-            d->outerContext->setContextObject(nullptr);
-
         if (d->hasVMEMetaObject || d->hasInterceptorMetaObject) {
             // This is somewhat dangerous because another thread might concurrently
             // try to resolve the dynamic metaobject. In practice this will then
@@ -255,8 +248,8 @@ void QQmlPrivate::qdeclarativeelement_destructor(QObject *o)
     }
 }
 
-QQmlData::QQmlData()
-    : ownMemory(true), indestructible(true), explicitIndestructibleSet(false),
+QQmlData::QQmlData(Ownership ownership)
+    : ownMemory(ownership == OwnsMemory), indestructible(true), explicitIndestructibleSet(false),
       hasTaintedV4Object(false), isQueuedForDeletion(false), rootObjectInCreation(false),
       hasInterceptorMetaObject(false), hasVMEMetaObject(false), hasConstWrapper(false),
       bindingBitsArraySize(InlineBindingArraySize), notifyList(nullptr),
@@ -312,7 +305,10 @@ void QQmlData::signalEmitted(QAbstractDeclarativeData *, QObject *object, int in
     // QQmlEngine to emit signals from a different thread.  These signals are then automatically
     // marshalled back onto the QObject's thread and handled by QML from there.  This is tested
     // by the qqmlecmascript::threadSignal() autotest.
-    if (!ddata->notifyList)
+
+    // Relaxed semantics here. If we're on a different thread we might schedule a useless event,
+    // but that should be rare.
+    if (!ddata->notifyList.loadRelaxed())
         return;
 
     auto objectThreadData = QObjectPrivate::get(object)->threadData.loadRelaxed();
@@ -386,11 +382,15 @@ int QQmlData::endpointCount(int index)
 
 void QQmlData::markAsDeleted(QObject *o)
 {
-    QQmlData::setQueuedForDeletion(o);
-
-    QObjectPrivate *p = QObjectPrivate::get(o);
-    for (QList<QObject *>::const_iterator it = p->children.constBegin(), end = p->children.constEnd(); it != end; ++it) {
-        QQmlData::markAsDeleted(*it);
+    QVarLengthArray<QObject *> workStack;
+    workStack.push_back(o);
+    while (!workStack.isEmpty()) {
+        auto currentObject = workStack.last();
+        workStack.pop_back();
+        QQmlData::setQueuedForDeletion(currentObject);
+        auto currentObjectPriv = QObjectPrivate::get(currentObject);
+        for (QObject *child: std::as_const(currentObjectPriv->children))
+            workStack.push_back(child);
     }
 }
 
@@ -400,9 +400,7 @@ void QQmlData::setQueuedForDeletion(QObject *object)
         if (QQmlData *ddata = QQmlData::get(object)) {
             if (ddata->ownContext) {
                 Q_ASSERT(ddata->ownContext.data() == ddata->context);
-                ddata->context->emitDestruction();
-                if (ddata->ownContext->contextObject() == object)
-                    ddata->ownContext->setContextObject(nullptr);
+                ddata->ownContext->deepClearContextObject(object);
                 ddata->ownContext.reset();
                 ddata->context = nullptr;
             }
@@ -413,7 +411,7 @@ void QQmlData::setQueuedForDeletion(QObject *object)
             // possible to get the metaobject anymore.
             // Also, there is no point in evaluating bindings in order to set properties on
             // half-deleted objects.
-            ddata->disconnectNotifiers();
+            ddata->disconnectNotifiers(DeleteNotifyList::No);
         }
     }
 }
@@ -437,34 +435,126 @@ void QQmlData::flushPendingBinding(int coreIndex)
 QQmlData::DeferredData::DeferredData() = default;
 QQmlData::DeferredData::~DeferredData() = default;
 
+template<>
+int qmlRegisterType<void>(const char *uri, int versionMajor, int versionMinor, const char *qmlName)
+{
+    QQmlPrivate::RegisterType type = {
+        QQmlPrivate::RegisterType::CurrentVersion,
+        QMetaType(),
+        QMetaType(),
+        0, nullptr, nullptr,
+        QString(),
+        nullptr,
+        uri,
+        QTypeRevision::fromVersion(versionMajor, versionMinor),
+        qmlName,
+        nullptr,
+        nullptr,
+        nullptr,
+        -1,
+        -1,
+        -1,
+        nullptr,
+        nullptr,
+        nullptr,
+        QTypeRevision::zero(),
+        -1,
+        QQmlPrivate::ValueTypeCreationMethod::None,
+    };
+
+    return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
+}
+
 bool QQmlEnginePrivate::baseModulesUninitialized = true;
 void QQmlEnginePrivate::init()
 {
     Q_Q(QQmlEngine);
 
     if (baseModulesUninitialized) {
+        // Named builtins
+        qmlRegisterType<void>("QML", 1, 0, "void");
+
+        const int varId = qmlRegisterType<QVariant>("QML", 1, 0, "var");
+        QQmlMetaType::registerTypeAlias(varId, QLatin1String("variant"));
+        qmlRegisterAnonymousSequentialContainer<QList<QVariant>>("QML", 1);
 
-        // required for the Compiler.
         qmlRegisterType<QObject>("QML", 1, 0, "QtObject");
         qmlRegisterType<QQmlComponent>("QML", 1, 0, "Component");
-        qmlRegisterAnonymousSequentialContainer<QList<QVariant>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<bool>>("QML", 1);
+
+        qmlRegisterType<int>("QML", 1, 0, "int");
         qmlRegisterAnonymousSequentialContainer<QList<int>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<float>>("QML", 1);
+
+        const int realId = qmlRegisterType<double>("QML", 1, 0, "real");
+        QQmlMetaType::registerTypeAlias(realId, QLatin1String("double"));
         qmlRegisterAnonymousSequentialContainer<QList<double>>("QML", 1);
+
+        qmlRegisterType<QString>("QML", 1, 0, "string");
         qmlRegisterAnonymousSequentialContainer<QList<QString>>("QML", 1);
-        qmlRegisterAnonymousSequentialContainer<QList<QUrl>>("QML", 1);
+
+        qmlRegisterType<bool>("QML", 1, 0, "bool");
+        qmlRegisterAnonymousSequentialContainer<QList<bool>>("QML", 1);
+
+        qmlRegisterType<QDateTime>("QML", 1, 0, "date");
         qmlRegisterAnonymousSequentialContainer<QList<QDateTime>>("QML", 1);
+
+        qmlRegisterType<QUrl>("QML", 1, 0, "url");
+        qmlRegisterAnonymousSequentialContainer<QList<QUrl>>("QML", 1);
+
+#if QT_CONFIG(regularexpression)
+        qmlRegisterType<QRegularExpression>("QML", 1, 0, "regexp");
         qmlRegisterAnonymousSequentialContainer<QList<QRegularExpression>>("QML", 1);
+#else
+        qmlRegisterType<void>("QML", 1, 0, "regexp");
+#endif
+
+        // Anonymous builtins
+        qmlRegisterAnonymousType<std::nullptr_t>("QML", 1);
+        qmlRegisterAnonymousType<QVariantMap>("QML", 1);
+
+        qmlRegisterAnonymousType<QJSValue>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<QJSValue>>("QML", 1);
+
+        qmlRegisterAnonymousType<qint8>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<qint8>>("QML", 1);
+
+        qmlRegisterAnonymousType<quint8>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<quint8>>("QML", 1);
+
+        qmlRegisterAnonymousType<short>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<short>>("QML", 1);
+
+        qmlRegisterAnonymousType<ushort>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<ushort>>("QML", 1);
+
+        qmlRegisterAnonymousType<uint>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<uint>>("QML", 1);
+
+        qmlRegisterAnonymousType<qlonglong>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<qlonglong>>("QML", 1);
+
+        qmlRegisterAnonymousType<qulonglong>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<qulonglong>>("QML", 1);
+
+        qmlRegisterAnonymousType<float>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<float>>("QML", 1);
+
+        qmlRegisterAnonymousType<QChar>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<QChar>>("QML", 1);
+
+        qmlRegisterAnonymousType<QDate>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<QDate>>("QML", 1);
+
+        qmlRegisterAnonymousType<QTime>("QML", 1);
+        qmlRegisterAnonymousSequentialContainer<QList<QTime>>("QML", 1);
+
+        qmlRegisterAnonymousType<QByteArray>("QML", 1);
         qmlRegisterAnonymousSequentialContainer<QList<QByteArray>>("QML", 1);
 
         // No need to specifically register those.
         static_assert(std::is_same_v<QStringList, QList<QString>>);
         static_assert(std::is_same_v<QVariantList, QList<QVariant>>);
 
-        qRegisterMetaType<QVariant>();
         qRegisterMetaType<QQmlScriptString>();
-        qRegisterMetaType<QJSValue>();
         qRegisterMetaType<QQmlComponent::Status>();
         qRegisterMetaType<QList<QObject*> >();
         qRegisterMetaType<QQmlBinding*>();
@@ -484,29 +574,16 @@ void QQmlEnginePrivate::init()
   \inmodule QtQml
   \brief The QQmlEngine class provides an environment for instantiating QML components.
 
-  Each QML component is instantiated in a QQmlContext.
-  QQmlContext's are essential for passing data to QML
-  components.  In QML, contexts are arranged hierarchically and this
-  hierarchy is managed by the QQmlEngine.
+  A QQmlEngine is used to manage \l{components}{QQmlComponent} and objects created from
+  them and execute their bindings and functions. QQmlEngine also inherits from
+  \l{QJSEngine} which allows seamless integration between your QML components and
+  JavaScript code.
 
-  Prior to creating any QML components, an application must have
-  created a QQmlEngine to gain access to a QML context.  The
-  following example shows how to create a simple Text item.
+  Each QML component is instantiated in a QQmlContext. In QML, contexts are arranged
+  hierarchically and this hierarchy is managed by the QQmlEngine. By default,
+  components are instantiated in the \l {QQmlEngine::rootContext()}{root context}.
 
-  \code
-  QQmlEngine engine;
-  QQmlComponent component(&engine);
-  component.setData("import QtQuick 2.0\nText { text: \"Hello world!\" }", QUrl());
-  QQuickItem *item = qobject_cast<QQuickItem *>(component.create());
-
-  //add item to view, etc
-  ...
-  \endcode
-
-  In this case, the Text item will be created in the engine's
-  \l {QQmlEngine::rootContext()}{root context}.
-
-  \sa QQmlComponent, QQmlContext, {QML Global Object}
+  \sa QQmlComponent, QQmlContext, {QML Global Object}, QQmlApplicationEngine
 */
 
 /*!
@@ -537,7 +614,7 @@ QQmlEngine::QQmlEngine(QQmlEnginePrivate &dd, QObject *parent)
   invalidated, but not destroyed (unless they are parented to the
   QQmlEngine object).
 
-  See QJSEngine docs for details on cleaning up the JS engine.
+  See ~QJSEngine() for details on cleaning up the JS engine.
 */
 QQmlEngine::~QQmlEngine()
 {
@@ -927,6 +1004,45 @@ void QQmlEngine::setOutputWarningsToStandardError(bool enabled)
     d->outputWarningsToMsgLog = enabled;
 }
 
+
+/*!
+  \since 6.6
+  If this method is called inside of a function that is part of
+  a binding in QML, the binding will be treated as a translation binding.
+
+  \code
+  class I18nAwareClass : public QObject {
+
+    //...
+
+     QString text() const
+     {
+          if (auto engine = qmlEngine(this))
+              engine->markCurrentFunctionAsTranslationBinding();
+          return tr("Hello, world!");
+     }
+  };
+  \endcode
+
+  \note This function is mostly useful if you wish to provide your
+  own alternative to the qsTr function. To ensure that properties
+  exposed from C++ classes are updated on language changes, it is
+  instead recommended to react to \c LanguageChange events. That
+  is a more general mechanism which also works when the class is
+  used in a non-QML context, and has slightly less overhead. However,
+  using \c markCurrentFunctionAsTranslationBinding can be acceptable
+  when the class is already closely tied to the QML engine.
+  For more details, see \l {Prepare for Dynamic Language Changes}
+
+  \sa QQmlEngine::retranslate
+*/
+void QQmlEngine::markCurrentFunctionAsTranslationBinding()
+{
+    Q_D(QQmlEngine);
+    if (auto propertyCapture = d->propertyCapture)
+        propertyCapture->captureTranslation();
+}
+
 /*!
   \internal
 
@@ -1222,49 +1338,73 @@ void QQmlData::releaseDeferredData()
 
 void QQmlData::addNotify(int index, QQmlNotifierEndpoint *endpoint)
 {
-    if (!notifyList) {
-        notifyList = (NotifyList *)malloc(sizeof(NotifyList));
-        notifyList->connectionMask = 0;
-        notifyList->maximumTodoIndex = 0;
-        notifyList->notifiesSize = 0;
-        notifyList->todo = nullptr;
-        notifyList->notifies = nullptr;
+    // Can only happen on "home" thread. We apply relaxed semantics when loading the atomics.
+
+    NotifyList *list = notifyList.loadRelaxed();
+
+    if (!list) {
+        list = new NotifyList;
+        // We don't really care when this change takes effect on other threads. The notifyList can
+        // only become non-null once in the life time of a QQmlData. It becomes null again when the
+        // underlying QObject is deleted. At that point any interaction with the QQmlData is UB
+        // anyway. So, for all intents and purposese, the list becomes non-null once and then stays
+        // non-null "forever". We can apply relaxed semantics.
+        notifyList.storeRelaxed(list);
     }
 
     Q_ASSERT(!endpoint->isConnected());
 
     index = qMin(index, 0xFFFF - 1);
-    notifyList->connectionMask |= (1ULL << quint64(index % 64));
 
-    if (index < notifyList->notifiesSize) {
+    // Likewise, we don't really care _when_ the change in the connectionMask is propagated to other
+    // threads. Cross-thread event ordering is inherently nondeterministic. Therefore, when querying
+    // the conenctionMask in the presence of concurrent modification, any result is correct.
+    list->connectionMask.storeRelaxed(
+            list->connectionMask.loadRelaxed() | (1ULL << quint64(index % 64)));
 
-        endpoint->next = notifyList->notifies[index];
+    if (index < list->notifiesSize) {
+        endpoint->next = list->notifies[index];
         if (endpoint->next) endpoint->next->prev = &endpoint->next;
-        endpoint->prev = &notifyList->notifies[index];
-        notifyList->notifies[index] = endpoint;
-
+        endpoint->prev = &list->notifies[index];
+        list->notifies[index] = endpoint;
     } else {
-        notifyList->maximumTodoIndex = qMax(int(notifyList->maximumTodoIndex), index);
+        list->maximumTodoIndex = qMax(int(list->maximumTodoIndex), index);
 
-        endpoint->next = notifyList->todo;
+        endpoint->next = list->todo;
         if (endpoint->next) endpoint->next->prev = &endpoint->next;
-        endpoint->prev = &notifyList->todo;
-        notifyList->todo = endpoint;
+        endpoint->prev = &list->todo;
+        list->todo = endpoint;
     }
 }
 
-void QQmlData::disconnectNotifiers()
+void QQmlData::disconnectNotifiers(QQmlData::DeleteNotifyList doDelete)
 {
-    if (notifyList) {
-        while (notifyList->todo)
-            notifyList->todo->disconnect();
-        for (int ii = 0; ii < notifyList->notifiesSize; ++ii) {
-            while (QQmlNotifierEndpoint *ep = notifyList->notifies[ii])
+    // Can only happen on "home" thread. We apply relaxed semantics when loading  the atomics.
+    if (NotifyList *list = notifyList.loadRelaxed()) {
+        while (QQmlNotifierEndpoint *todo = list->todo)
+            todo->disconnect();
+        for (int ii = 0; ii < list->notifiesSize; ++ii) {
+            while (QQmlNotifierEndpoint *ep = list->notifies[ii])
                 ep->disconnect();
         }
-        free(notifyList->notifies);
-        free(notifyList);
-        notifyList = nullptr;
+        free(list->notifies);
+
+        if (doDelete == DeleteNotifyList::Yes) {
+            // We can only get here from QQmlData::destroyed(), and that can only come from the
+            // the QObject dtor. If you're still sending signals at that point you have UB already
+            // without any threads. Therefore, it's enough to apply relaxed semantics.
+            notifyList.storeRelaxed(nullptr);
+            delete list;
+        } else {
+            // We can use relaxed semantics here. The worst thing that can happen is that some
+            // signal is falsely reported as connected. Signal connectedness across threads
+            // is not quite deterministic anyway.
+            list->connectionMask.storeRelaxed(0);
+            list->maximumTodoIndex = 0;
+            list->notifiesSize = 0;
+            list->notifies = nullptr;
+
+        }
     }
 }
 
@@ -1349,7 +1489,7 @@ void QQmlData::destroyed(QObject *object)
             guard->objectDestroyed(guard);
     }
 
-    disconnectNotifiers();
+    disconnectNotifiers(DeleteNotifyList::Yes);
 
     if (extendedData)
         delete extendedData;
@@ -1390,7 +1530,7 @@ QQmlData *QQmlData::createQQmlData(QObjectPrivate *priv)
 {
     Q_ASSERT(priv);
     Q_ASSERT(!priv->isDeletingChildren);
-    priv->declarativeData = new QQmlData;
+    priv->declarativeData = new QQmlData(OwnsMemory);
     return static_cast<QQmlData *>(priv->declarativeData);
 }
 
@@ -1545,7 +1685,8 @@ void QQmlEnginePrivate::cleanupScarceResources()
 
   The newly added \a path will be first in the importPathList().
 
-  \sa setImportPathList(), {QML Modules}, {QML Import Path}
+  \b {See also} \l setImportPathList(), \l {QML Modules},
+    and \l [QtQml] {QML Import Path}
 */
 void QQmlEngine::addImportPath(const QString& path)
 {
@@ -2033,10 +2174,12 @@ LoadHelper::LoadHelper(QQmlTypeLoader *loader, QAnyStringView uri)
 
 {
     auto import = std::make_shared<PendingImport>();
-    import->uri = uri.toString();
+    import->uri = m_uri;
     QList<QQmlError> errorList;
-    if (!Blob::addImport(import, &errorList))
-        m_uri = QString(); // reset m_uri to remember the failure
+    if (!Blob::addImport(import, &errorList)) {
+        qCDebug(lcQmlImport) << "LoadHelper: Errors loading " << m_uri << errorList;
+        m_uri.clear(); // reset m_uri to remember the failure
+    }
 }
 
 LoadHelper::ResolveTypeResult LoadHelper::resolveType(QAnyStringView typeName)
