@@ -19,6 +19,7 @@
 #include <QtQuickTestUtils/private/qmlutils_p.h>
 #include <QtQuickTestUtils/private/visualtestutils_p.h>
 #include <QtQuickTestUtils/private/viewtestutils_p.h>
+#include <QtQuick/private/qquickmousearea_p.h>
 
 #include <QtGui/private/qeventpoint_p.h>
 
@@ -140,6 +141,8 @@ private slots:
     void hoverPropagation_siblings();
     void hoverEnterOnItemMove();
     void hoverEnterOnItemMoveAfterHide();
+    void clearItemsOnHoverLeave();
+    void deleteTargetOnPress();
 
 private:
     QScopedPointer<QPointingDevice> touchDevice = QScopedPointer<QPointingDevice>(QTest::createTouchDevice());
@@ -177,7 +180,8 @@ void tst_qquickdeliveryagent::passiveGrabberOrder()
 
     QPoint pos(75, 75);
     QTest::mousePress(&view, Qt::LeftButton, Qt::NoModifier, pos);
-    QTest::qWait(1000);
+    QTRY_VERIFY(rootTap->isPressed());
+    QTRY_VERIFY(subsceneTap->isPressed());
     auto devPriv = QPointingDevicePrivate::get(QPointingDevice::primaryPointingDevice());
     const auto &persistentPoint = devPriv->activePoints.values().first();
     qCDebug(lcTests) << "passive grabbers" << persistentPoint.passiveGrabbers << "contexts" << persistentPoint.passiveGrabbersContext;
@@ -187,7 +191,8 @@ void tst_qquickdeliveryagent::passiveGrabberOrder()
     QCOMPARE(persistentPoint.passiveGrabbers.last(), rootTap);
 
     QTest::mouseRelease(&view, Qt::LeftButton);
-    QTest::qWait(100);
+    QTRY_COMPARE(rootTap->isPressed(), false);
+    QTRY_COMPARE(subsceneTap->isPressed(), false);
     // QQuickWindow::event() has failsafe: clear all grabbers after release
     QCOMPARE(persistentPoint.passiveGrabbers.size(), 0);
 
@@ -582,6 +587,47 @@ void tst_qquickdeliveryagent::hoverEnterOnItemMoveAfterHide()
     QCOMPARE(hoverItem.hoverEnter, false);
 }
 
+void tst_qquickdeliveryagent::clearItemsOnHoverLeave()
+{
+    QQuickView window;
+    QVERIFY(QQuickTest::showView(window, testFileUrl("clearItemsOnHoverLeave.qml")));
+
+    QTest::mouseMove(&window, QPoint(10, 205)); // Move to MouseArea that triggers close
+    QTest::mouseMove(&window, QPoint(10, 405)); // Exit MouseArea that triggers close.
+}
+
+// QTBUG-91272
+void tst_qquickdeliveryagent::deleteTargetOnPress()
+{
+    QQuickWindow window;
+    auto deliveryAgent = QQuickWindowPrivate::get(&window)->deliveryAgentPrivate();
+    window.resize(200, 200);
+    window.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&window));
+
+    QQuickMouseArea *lowerArea = new QQuickMouseArea(window.contentItem());
+    lowerArea->setWidth(200);
+    lowerArea->setHeight(200);
+
+    QQuickMouseArea *upperArea = new QQuickMouseArea(window.contentItem());
+    upperArea->setWidth(180);
+    upperArea->setHeight(180);
+    bool pressed = false;
+    connect(upperArea, &QQuickMouseArea::pressed, this, [&]() {
+        pressed = true;
+        delete lowerArea;
+        lowerArea = nullptr;
+    });
+    QTest::mouseMove(&window, QPoint(100, 100));
+    QTest::mousePress(&window, Qt::MouseButton::LeftButton, {}, {100, 100});
+    deliveryAgent->flushFrameSynchronousEvents(&window);
+    QVERIFY(pressed);
+    QVERIFY(upperArea->isPressed());
+    QTest::mouseRelease(&window, Qt::MouseButton::LeftButton, {}, {100, 100});
+    deliveryAgent->flushFrameSynchronousEvents(&window);
+    QVERIFY(!upperArea->isPressed());
+}
+
 QTEST_MAIN(tst_qquickdeliveryagent)
 
 #include "tst_qquickdeliveryagent.moc"
